<?php

/**
 * Copyright 2019 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Entities;

use JsonSerializable;

/**
 * DPA version entity
 */
final class Dpa implements JsonSerializable {

	/**
	 * @var OsDpaAttributtes Attributes
	 */
	private $attributes;

	/**
	 * @var string Download path
	 */
	private $downloadPath;

	/**
	 * @var string DPA version
	 */
	private $version;

	/**
	 * Constructor
	 * @param string $version DPA version
	 * @param OsDpaAttributtes $attributes Attributes
	 * @param string $downloadPath Download path
	 */
	public function __construct(string $version, OsDpaAttributtes $attributes, string $downloadPath) {
		$this->attributes = $attributes;
		$this->downloadPath = $downloadPath;
		$this->version = $version;
	}

	/**
	 * Creates a new DPA entity from an array
	 * @param array<string, string|int> $data JSON data in an array
	 * @return Dpa Created DPA entity
	 */
	public static function fromArray(array $data): self {
		return new self($data['version'], OsDpaAttributtes::fromInt($data['attributes']), $data['downloadPath']);
	}

	/**
	 * Returns DPA version attributes
	 * @return OsDpaAttributtes DPA version attributes
	 */
	public function getAttributes(): OsDpaAttributtes {
		return $this->attributes;
	}

	/**
	 * Returns the download path
	 * @return string Download path
	 */
	public function getDownloadPath(): string {
		return $this->downloadPath;
	}

	/**
	 * Returns the DPA version
	 * @param bool $pretty Pretty formatting
	 * @return string DPA version
	 */
	public function getVersion(bool $pretty = false): string {
		if ($pretty) {
			$versionInt = hexdec($this->version);
			$major = strval($versionInt >> 8);
			$minor = str_pad(dechex($versionInt & 0xff), 2, '0', STR_PAD_LEFT);
			return $major . '.' . $minor;
		}
		return $this->version;
	}

	/**
	 * Serialized the DPA entity into JSON
	 * @return array<string, string|array<string, bool|null>> JSON serialized DPA entity
	 */
	public function jsonSerialize(): array {
		return [
			'version' => $this->getVersion(true),
			'attributes' => $this->attributes->jsonSerialize(),
			'downloadPath' => $this->downloadPath,
		];
	}

}
